package net.pingex.dcf.commands;

import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

/**
 * Represents a single command.
 */
public abstract class Command implements ICommandExecutor
{
	/**
	 * Main name of the command
	 */
	private String name;

	/**
	 * Command can also be called using the following list of aliases
	 */
	private List<String> aliases;

	/**
	 * Description of the command.
	 */
	private String description;

	/**
	 * Is the command enabled ? Can it be invoked ?
	 */
	private boolean isEnabled;

	/**
	 * Command usage help
	 */
	private String usage;

	/**
	 * Basic constructor.
	 * @param name Name of the command
	 * @param aliases Aliases, if any
	 * @param description Description of the command
	 * @param isEnabled Is the command enabled ?
	 * @param usage Command usage help
	 */
	public Command(String name, List<String> aliases, String description, boolean isEnabled, String usage)
	{
		this.name = name;
		this.aliases = aliases;
		this.description = description;
		this.isEnabled = isEnabled;
		this.usage = usage;
	}

	/**
	 * Contains Command default values
	 */
	public static class Defaults
	{
		public static final List<String> ALIASES = Collections.emptyList();
		public static final String DESCRIPTION = "No command description provided.";
		public static final boolean IS_ENABLED = true;
		public static final String USAGE = "No command usage provided.";
	}

	/**
	 * Builder for a command
	 */
	public static class Builder
	{
		/**
		 * Main name of the command
		 */
		private String name;

		/**
		 * Command can also be called using the following list of aliases
		 */
		private List<String> aliases = Defaults.ALIASES;

		/**
		 * Description of the command.
		 */
		private String description = Defaults.DESCRIPTION;

		/**
		 * Is the command enabled ? Can it be invoked ?
		 */
		private boolean isEnabled = Defaults.IS_ENABLED;

		/**
		 * Command usage help
		 */
		private String usage = Defaults.USAGE;

		public Builder(String name)
		{
			this.name = name;
		}

		public Builder aliases(List<String> aliases)
		{
			this.aliases = aliases;
			return this;
		}

		/**
		 * Set aliases using varargs
		 * @param aliases Aliases varargs
		 */
		public Builder aliases(String... aliases)
		{
			this.aliases = Arrays.asList(aliases);
			return this;
		}

		public Builder description(String description)
		{
			this.description = description;
			return this;
		}

		public Builder enabled(boolean enabled)
		{
			isEnabled = enabled;
			return this;
		}

		public Builder usage(String usage)
		{
			this.usage = usage;
			return this;
		}

		/**
		 * Build a new Command using a supplied executor.
		 * @param toExecute The body of the command.
		 * @return Built command.
		 */
		public Command build(ICommandExecutor toExecute)
		{
			return new Command(name, aliases, description, isEnabled, usage)
			{
				@Override
				public void execute(List<String> arguments) throws Throwable
				{
					toExecute.execute(arguments);
				}
			};
		}

		/**
		 * Build a new Command using a method and a target object (for use with annotated command)
		 * @param target The method which needs to be invoked
		 * @param invokable The matching object
		 * @return Built command.
		 */
		public Command build(Method target, Object invokable)
		{
			return build(arguments -> target.invoke(invokable));
		}
	}

	/**
	 * Get a builder for a new command.
	 * @param name The name of the future command
	 * @return A new manipulable builder to build the command.
	 */
	public static Builder builder(String name)
	{
		return new Builder(name);
	}

	/**
	 * Create a new command object from an annotated method
	 * @param target Target method
	 * @return A new created command built from the method, or `null` if the method is not a valid command.
	 */
	public static Command buildFromAnnotatedCommand(AnnotatedCommand meta, Method target, Object invokable)
	{
		return builder(meta.name())
				.aliases(Arrays.asList(meta.aliases()))
				.description(meta.description())
				.enabled(meta.isEnabled())
				.usage(meta.usage())
				.build(target, invokable);
	}

	public String getName()
	{
		return name;
	}

	public List<String> getAliases()
	{
		return aliases;
	}

	public String getDescription()
	{
		return description;
	}

	public boolean isEnabled()
	{
		return isEnabled;
	}

	public String getUsage()
	{
		return usage;
	}

	public void setEnabled(boolean enabled)
	{
		isEnabled = enabled;
	}
}
